#include "Forecast_Util.h"

using namespace Forecast_Util;

/*
 * ラプラシアン の差分式の定義
 */
double Forecast_Util::L(Param_List& pm, GMatrix& mx, int i, int j)
{
    // 引数で与えられたデータの(i, j) における,ラプラシアンを計算する
    double tmp_x, tmp_y;
    double epsilon2 = (pm.del_x * pm.del_x) / ( pm.del_y * pm.del_y);

    // x 方向に関しては, 周期境界条件を適用する
    if(i == 0){
        tmp_x = mx.Get(i+1, j) + mx.Get(pm.num_x-1, j) - 2.0 * mx.Get(i, j);

    }else if(i == pm.num_x-1){
        tmp_x = mx.Get(0, j) + mx.Get(i-1, j) - 2.0 * mx.Get(i, j);
    
    }else{
        tmp_x = mx.Get(i+1, j) + mx.Get(i-1, j) - 2.0 * mx.Get(i, j);
    
    }
    
    // y 方向の境界に関しては,固定境界である
    if(j == 0 || j == pm.num_y-1){
        tmp_y = 0.0;
    }else{
        tmp_y = epsilon2 *  ( mx.Get(i, j +1) + mx.Get(i, j-1) - 2.0 * mx.Get(i, j) );
    }

    return tmp_x + tmp_y;
}


/*
 * ヤコビアン の差分式の定義
 */   
double Forecast_Util::J(Param_List& pm, GMathLib::GMatrix& mx1, GMathLib::GMatrix& mx2, int i, int j)
{

    // 引数で与えられたデータの(i, j) における,ヤコビアンを計算する
    //J = tmp1 * tmp2 - tmp3 * tmp4
    double tmp1, tmp2, tmp3, tmp4;

    // x 方向に関しては周期境界条件
    if(i == 0){
         tmp1 = mx1.Get(1, j) - mx1.Get(pm.num_x - 1, j);
         tmp4 = mx2.Get(1, j) - mx2.Get(pm.num_x - 1, j);
    
    }else if(i == pm.num_x-1){
         tmp1 = mx1.Get(0, j) - mx1.Get(i-1, j);
         tmp4 = mx2.Get(0, j) - mx2.Get(i-1, j);
    
    }else{
         tmp1 = mx1.Get(i+1, j) - mx1.Get(i-1, j);
         tmp4 = mx2.Get(i+1, j) - mx2.Get(i-1, j);
    }

    // y 方向の境界に関しては,固定端境界である
    if(j == 0){
         tmp2 = 2.0 * ( mx2.Get(i, 1) - mx2.Get(i, 0) );
         tmp3 = 2.0 * ( mx1.Get(i, 1) - mx1.Get(i, 0) );
    
    }else if(j == pm.num_y-1){
         tmp2 = 2.0 * ( mx2.Get(i, j) - mx2.Get(i, j-1) );
         tmp3 = 2.0 * ( mx1.Get(i, j) - mx1.Get(i, j-1) );

    }else{
         tmp2 = mx2.Get(i, j+1) - mx2.Get(i, j-1);
         tmp3 = mx1.Get(i, j+1) - mx1.Get(i, j-1);

    }

    return tmp1 * tmp2 - tmp3 * tmp4;

}

void Forecast_Util::solve_ODE_Direct(Param_List& pm, GMatrix* p_A, GVector* p_mean_mx, GVector* p_eval_mx, double A, double B)
{
     // d^2 f(y) / d y^2 + h * f(y) = r(y)
     // A ( f_y+1 + f_y-1 - 2 * f_y ) + B f_y = r_y
     // f'(0) = f'(N) = 0
     
     GLinear_Homo_Eq solver;
     double jmax = p_mean_mx->Size();
     double tmp_jj = - 2.0 * A + B;

     for(int i=0; i < jmax; i++){
         for(int j=0; j < jmax; j++){
	     if(j == i){
                 p_A->Set(i, j, tmp_jj);
	     }else if(j == i-1 || j == i+1){
                 p_A->Set(i, j, A);
	     }
	 }
     }

     p_A->Set(0, 0, 1.0);
     p_A->Set(0, 1, -1.0);
     p_A->Set(1, 0, - A);
     p_A->Set(1, 1, B);
     p_A->Set(jmax-2, jmax-2, B);
     p_A->Set(jmax-2, jmax-1, - A);
     p_A->Set(jmax-1, jmax-2, -1.0);
     p_A->Set(jmax-1, jmax-1, 1.0);

     p_eval_mx->Set(0, 0.0);
     p_eval_mx->Set(jmax-1, 0.0);

     solver.S_Gauss_Partial_Pivot(*p_A, *p_eval_mx, *p_mean_mx);
     
}

void Forecast_Util::solve_Poisson_Eq_Relax(Param_List& pm, GMatrix* p_V, GMatrix* p_S, double A, double B, double conv_value, double v)
{
     // [ A L_ij + B ] V_ij = S_ij 
     int imax = p_V->Row();
     int jmax = p_V->Column();
     int i_up, i_dwn;
     double tmp;
     
     double ep2 = ( pm.del_x * pm.del_x ) / ( pm.del_y * pm.del_y );
     double conv_max = 0.0;
     double check_convergence;
     int MAX_COUNTER = 300;//1e05;
     int counter = 0;

     // 250 hPa あるいは 750 hPa の del_t 秒後のポテンシャル場を求めるために,
     // 差分方程式を緩和法で解く.
     while(counter < MAX_COUNTER){
         for(int i=0; i < imax; i++){
             
	     // 東西方向に対して周期境界条件を適用する.
	     if(i == 0){ 
	         i_dwn = imax-1; i_up = 1;
	     }else if(i == imax-1){
	         i_up = 0; i_dwn = imax-2;
	     }else{
	         i_dwn = i-1; i_up = i+1;
	     }

	     for(int j=1; j < jmax-1; j++){
                  tmp = p_V->Get(i, j) + v * 
			 (
			   (
	                     A * (  
	                      p_V->Get(i_up, j) + p_V->Get(i_dwn, j) + ep2 * ( p_V->Get(i, j+1) + p_V->Get(i, j-1) )
		             )
		             - p_S->Get(i, j)
		           )  
			   / ( 2.0 * A * ( 1 + ep2 ) - B )
			   - p_V->Get(i, j)
			 );
                  
		  check_convergence = std::abs(tmp - p_V->Get(i, j));
		  if(check_convergence > conv_max){
		      conv_max = check_convergence;
		  }
                  
		  p_V->Set(i, j, tmp);
	     }
         }

	 // 全領域において, conv_max が収束判定閾値の　conv_value [m^2/s] 以下になったらループを抜ける
	 if(conv_max < conv_value){
	     //std::cout << counter <<",";
	     break;
	 }

	 if(counter == MAX_COUNTER-1){
	     std::cout << "waring with reaxation method: solution is Not expected convergent." << std::endl;
	     std::cout << "conv_value: " << conv_value << ", maxmum error:" << conv_max << std::endl;
	 }

         counter++;
	 conv_max = 0.0;
     }
}

void Forecast_Util::solve_Poisson_Eq_Relax2(Param_List& pm, GMatrix* p_SF1, GMatrix* p_SF3, GMatrix* p_PV1, GMatrix* p_PV3, double A, double B, double conv_value, double v)
{
     int imax = p_SF1->Row();
     int jmax = p_SF1->Column();
     int i_up, i_dwn;
     double tmp;
     
     double ep2 = ( pm.del_x * pm.del_x ) / ( pm.del_y * pm.del_y );
     double conv_max1 = 0.0;
     double conv_max3 = 0.0;
     double check_convergence1;
     double check_convergence3;

     int MAX_COUNTER = 300;//1e05;
     int counter = 0;

     // 250 hPa あるいは 750 hPa の del_t 秒後のポテンシャル場を求めるために,
     // 差分方程式を緩和法で解く.
     while(counter < MAX_COUNTER){
         for(int i=0; i < imax; i++){
             
	     // 東西方向に対して周期境界条件を適用する.
	     if(i == 0){ 
	         i_dwn = imax-1; i_up = 1;
	     }else if(i == imax-1){
	         i_up = 0; i_dwn = imax-2;
	     }else{
	         i_dwn = i-1; i_up = i+1;
	     }

	     for(int j=1; j < jmax-1; j++){
                  tmp = p_SF1->Get(i, j) + v * 
			(
			  (
	                    A * (  
	                     p_SF1->Get(i_up, j) + p_SF1->Get(i_dwn, j) + ep2 * ( p_SF1->Get(i, j+1) + p_SF1->Get(i, j-1) )
		            )
		            - p_PV1->Get(i, j) + B * p_SF3->Get(i, j) 
		          )  / ( 2.0 * A * ( 1 + ep2 ) + B )
			  - p_SF1->Get(i, j)
			);
                  
		  check_convergence1 = std::abs(tmp - p_SF1->Get(i, j));
		  if(check_convergence1 > conv_max1){
		      conv_max1 = check_convergence1;
		  }
                  
		  p_SF1->Set(i, j, tmp);
                  
		  tmp = p_SF3->Get(i, j) + v * 
			(
			  (
	                    A * (  
	                     p_SF3->Get(i_up, j) + p_SF3->Get(i_dwn, j) + ep2 * ( p_SF3->Get(i, j+1) + p_SF3->Get(i, j-1) )
		            )
		            - p_PV3->Get(i, j) + B * p_SF1->Get(i, j)  
		          )  / ( 2.0 * A * ( 1 + ep2 ) + B )
			  - p_SF3->Get(i, j)
			);
                  
		  check_convergence3 = std::abs(tmp - p_SF3->Get(i, j));
		  if(check_convergence3 > conv_max3){
		      conv_max3 = check_convergence3;
		  }
                  
		  p_SF3->Set(i, j, tmp);
	     }
         }

	 // 全領域において, conv_max が収束判定閾値の　conv_value [m^2/s] 以下になったらループを抜ける
	 if( conv_max1 + conv_max3 < conv_value){
	     //std::cout << counter <<",";
	     break;
	 }

	 if(counter == MAX_COUNTER-1){
	     std::cout << "waring with reaxation method: solution is Not expected convergent." << std::endl;
	     std::cout << "conv_value: " << conv_value << ", maxmum error:" << conv_max1 + conv_max3 << std::endl;
	 }

         counter++;
	 conv_max1 = 0.0;
	 conv_max3 = 0.0;
     }
}


/**
 * 行列データの東西平均を取る関数
 */
void Forecast_Util::zonal_average(GMatrix& src_mx, GVector& avg_mx){
    int imax = src_mx.Row();
    double tmp = 0.0;

    for(int j=0; j < src_mx.Column(); j++){ 
        for(int i=0; i < imax; i++){
	     tmp += src_mx.Get(i, j);
	}

        avg_mx.Set(j, tmp / imax);
	tmp = 0.0;
    }

}

void Forecast_Util::calc_disturb(GMatrix& src_mx, GVector& mean_mx, GMatrix& distb_mx)
{
    for(int i=0; i < src_mx.Row(); i++){
        for(int j=0; j < src_mx.Column(); j++){
             distb_mx.Set(i, j, src_mx.Get(i, j) - mean_mx.Get(j));
	}
    }
}

void Forecast_Util::Add_Mean_Disturb(GVector& mean_mx, GMatrix& distb_mx, GMatrix& store_mx)
{
     double tmp;

     for(int j=0; j < store_mx.Column(); j++){
         tmp = mean_mx.Get(j);

         for(int i=0; i < store_mx.Row(); i++){
	     store_mx.Set(i, j, tmp + distb_mx.Get(i, j));
	 }
      }
}

